from typing import Tuple, List, Any, Union
import os
import json
import logging

import h5py
import matplotlib.pyplot as plt
import numpy as np
from .._pipeline import Pipeline
from ...analysis.mean_squared_displacement import MeanSquaredDisplacement
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import Simulation, GroupIterator

logger = logging.getLogger(__name__)


class BeadMSD(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 overwrite: bool = False):
        super().__init__(experiment_configuration)
        self.skip = experiment_configuration.skip_n_frames
        self.plot_files.update({
            'MSD': os.path.join(experiment_configuration['MSD'], 'ensemble_msds.svg'),
        })
        self.output_files.update({
            'MSD': os.path.join(experiment_configuration['MSD'], 'msd_ensemble.h5'),
        })

        if overwrite:
            self._remove_output_files()

    def run_analysis(self) -> Tuple[plt.Figure, plt.Axes]:
        msd_data = self.get_ensemble_msds()
        fig, ax = self._plot_msds(msd_data)
        return fig, ax

    def _validate_configuration(self):
        assert "MSD" in self.experiment_configuration

    def get_ensemble_msds(self) -> 'EnsembleMSDData':
        iterator = self.experiment_configuration.experiment_iterator
        data = EnsembleMSDData()
        for group in iterator:
            selected_simulations = [sim for sim in group]
            msd_ensemble = self._get_ensemble_msds_group(group, selected_simulations)

            radius, bead_diffusion_const = self._load_bead_diffusion_const(selected_simulations[0])
            data.append_data(
                msd_ensemble.lag_time,
                msd_ensemble.msd,
                radius,
                bead_diffusion_const,
                group.latex_label,
                group.color
            )
        return data

    def _plot_msds(self, data: 'EnsembleMSDData') -> Tuple[plt.Figure, plt.Axes]:
        filename = self.plot_files['MSD']
        fig, ax = data.plot(filename)
        return fig, ax

    def _load_bead_diffusion_const(self, sim: Simulation) -> Tuple[float, float]:
        dr = sim.data_reader
        if hasattr(dr, 'read_bead_diffusion_constant') and hasattr(dr, 'read_bead_radius'):
            return dr.read_bead_radius(), dr.read_bead_diffusion_constant()
        init_idx = dr.read_parameter('initial_state_index')
        format_string = BeadMSD._get_network_config_format_string(self.experiment_configuration.root_folder)
        init_config = format_string.format(init_idx)
        with open(init_config, 'rt') as fh:
            radius = json.load(fh)['acceptance_rate_handler']['bead_radius']
            bead_diffusion_const = 0.5 / radius
        return radius, bead_diffusion_const

    def _get_ensemble_msds_group(self, group: GroupIterator,
                                 selected_simulations) -> MeanSquaredDisplacement:
        h5_group_template = group.get_label_from_values() + f"_skip_{self.skip}"
        h5_group_name = h5_group_template.format(**group.values)
        filename = self.output_files['MSD']
        lag_time, msd_ensemble, var_ensemble = BeadMSD._load_ensemble_msds(filename, h5_group_name)
        if lag_time is not None:
            logger.info(f"Loaded MSD data of group {group.label} from file.")
            return MeanSquaredDisplacement(np.arange(1, len(lag_time)+1), msd_ensemble, var_ensemble, lag_time)

        logger.info(f"Computing MSD data of {group.label}, will save to file for next use.")
        msds = self._get_individual_msds(selected_simulations)
        lag_time = msds[0].lag_time
        msd_ensemble = self._get_mean_of_individuals(msds)
        BeadMSD._write_ensemble_msds(filename, h5_group_name, lag_time, msd_ensemble.msd, msd_ensemble.variance_of_msd)
        return msd_ensemble

    @staticmethod
    def _get_mean_of_individuals(msds: List[MeanSquaredDisplacement]) -> MeanSquaredDisplacement:
        mean_msd = np.full((len(msds), len(msds[0].msd)), np.nan)

        for i in range(len(msds)):
            mean_msd[i] = msds[i].msd

        var_msd = np.nanvar(mean_msd, axis=0)
        mean_msd = np.nanmean(mean_msd, axis=0)

        return MeanSquaredDisplacement(msds[0].lag, mean_msd, var_msd, msds[0].lag_time)

    @staticmethod
    def _write_ensemble_msds(filename, h5_group_name, lag_time, msd_ensemble, var_ensemble):
        with h5py.File(filename, 'a') as msds_file:
            h5_group = msds_file.create_group(h5_group_name)
            h5_group['log_msd'] = msd_ensemble
            h5_group['variance_of_msd'] = var_ensemble
            h5_group['log_lag_time'] = lag_time

    @staticmethod
    def _load_ensemble_msds(
            filename: str,
            h5_group_name: str
    ) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[None, None, None]]:
        if not os.path.exists(filename):
            return None, None, None
        with h5py.File(filename, 'r') as msds_file:
            if h5_group_name not in msds_file:
                return None, None, None
            h5_group = msds_file[h5_group_name]
            if 'log_msd' not in h5_group:
                return None, None, None
            if 'variance_of_msd' not in h5_group:
                return None, None, None
            return h5_group['log_lag_time'][:], h5_group['log_msd'][:], h5_group['variance_of_msd'][:]

    def _get_individual_msds(
            self, simulations: List[Simulation]
    ) -> List[MeanSquaredDisplacement]:
        lag_time = None
        msds = []
        for sim in simulations:
            a = sim.analyser
            n_non_filament = a.get_offset_filaments()
            msds_sim = a.get_msd(list(range(n_non_filament)), non_filament=True, skip=self.skip)

            t = a.get_time()
            delta_t = t[1] - t[0]
            lags = msds_sim[0].lag
            if lag_time is None:
                lag_time = lags * delta_t
            else:
                assert (lag_time == lags * delta_t).all()
            for msd_i in msds_sim:
                msd_i.lag_time = lag_time
            msds += msds_sim
        return msds

    @staticmethod
    def _get_network_config_format_string(root_folder: str) -> str:
        for i in range(6):
            guessed_folder_name = os.path.join(root_folder, 'initial_states', f'network{"0" * i}')
            if os.path.exists(guessed_folder_name):
                return os.path.join(root_folder, 'initial_states', 'network{:0' + str(i) + '}',
                                    'config.json')
        raise RuntimeError("Unable to guess folder name of initial network, tried initial_states/"
                           "network0 to initial_states/network000000. This algorithm checks only for up to"
                           " 6 digits.")


class EnsembleMSDData:

    def __init__(self):
        self.lag_times = []
        self.msds = []
        self.bead_radii = []
        self.bead_diffusion_constants = []
        self.labels = []
        self.colors = []

    def append_data(self, lag_time, msd, bead_radius, bead_diffusion_constant, label, color):
        self.lag_times.append(lag_time)
        self.msds.append(msd)
        self.bead_radii.append(bead_radius)
        self.bead_diffusion_constants.append(bead_diffusion_constant)
        self.labels.append(label)
        self.colors.append(color)

    def get_lag_time_and_msd(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        return self.lag_times[index], self.msds[index]

    @property
    def length(self) -> int:
        return len(self.lag_times)

    def plot(self, filename: str) -> Tuple[plt.Figure, plt.Axes]:
        fig, ax = plt.subplots(1, 1)

        global_max = 0
        for i in range(self.length):
            msd = self.msds[i]
            global_max = max(global_max, msd.max())
            ax.plot(
                self.lag_times[i],
                msd,
                color=self.colors[i],
                label=self.labels[i]
            )
        colors, unique_free_diffusions = self._get_unique_free_diffusions()
        for i, (c, u) in enumerate(zip(colors, unique_free_diffusions)):
            label = 'free diff.' if i == 0 else None
            ax.plot(self.lag_times[i], u, '--', color=c,
                    label=label, zorder=1, alpha=0.7)

        ax.set(
            xscale='log',
            yscale='log',
            xlabel='lag time $\\tau / t_0$',
            ylabel='MSD $\\langle \\Delta r^2 \\rangle / x_0^2$',
        )
        # ax.legend(bbox_to_anchor=[1.05, 1.0], loc='upper left')
        ax.legend(
            loc='upper left',  # bbox_to_anchor=(0.5, 1.05),
            ncol=2
        )
        ylim = ax.get_ylim()
        ax.set_ylim(ylim[0], global_max * 1.2)
        fig.tight_layout()
        fig.savefig(filename)
        return fig, ax

    def _get_unique_free_diffusions(self) -> Tuple[List[Any], List[np.ndarray]]:
        diff_consts = np.array(self.bead_diffusion_constants)
        uniques = np.unique(diff_consts)
        if len(uniques) == 1:
            return ['black'], [6 * self.lag_times[0] * uniques[0]]
        elif len(uniques) == self.length:
            return (self.colors,
                    [6 * self.lag_times[i] * self.bead_diffusion_constants[i] for i in range(self.length)])
        raise RuntimeError("Can only handle cases where all beads have same size or all beads have different size."
                           f" Your experiment has {self.length} ensemble MSDs "
                           f"and {len(uniques)} different bead sizes.")



from typing import List, Tuple, Union, Dict, Any, Type, Generator, Optional
from dataclasses import dataclass
import os
import pandas as pd
import numpy as np
from matplotlib.colors import Colormap, Normalize
from actomyosin_analyser.file_io.data_reader import DataReader
from ..analysis.analyser import Analyser


class ExperimentIterator:

    def __init__(self, data_reader_class: Type[DataReader],
                 group_by: List[Tuple[pd.DataFrame, List[str]]],
                 simulation_folder_template: str, exclude: List[int],
                 data_file_suffix: str='data.h5'):

        index0 = group_by[0][0].index
        previous_groups = [GroupIterator(data_reader_class, index0, {}, simulation_folder_template)]
        for table, columns in group_by:
            assert (table.index == index0).all()
            for col in columns:
                _groups = []
                for group in previous_groups:
                    sim_indices = np.array([idx for idx in group.index if idx not in exclude])
                    if len(sim_indices) == 0:
                        continue
                    selected = table.loc[sim_indices]
                    for value in selected[col].unique():
                        new_values = group.values.copy()
                        new_values.update({col: value})
                        new_group = GroupIterator(
                            data_reader_class=data_reader_class,
                            index=selected.index[selected[col] == value],
                            values=new_values,
                            simulation_folder_template=simulation_folder_template,
                            data_file_suffix=data_file_suffix
                        )
                        _groups.append(new_group)
                previous_groups = _groups
        self.groups = previous_groups

    def __iter__(self):
        return iter(self.groups)

    def __len__(self):
        return len(self.groups)

    def get(self, **kwargs) -> List['GroupIterator']:
        selected = []
        for g in self.groups:
            if not g.has_matching_values(**kwargs):
                continue
            selected.append(g)
        return selected

    def create_index_from_groups(self) -> pd.Index:
        values = self.groups[0].values
        names = list(values.keys())
        arrays = []
        for n in names:
            arrays.append([g.values[n] for g in self.groups])
        if len(names) == 1:
            return pd.Index(arrays[0], name=names[0])
        return pd.MultiIndex.from_arrays(arrays, names=names)

    def assign_group_colors(
            self,
            key: str,
            colormap: Colormap,
            by_order: bool=True,
            norm: Optional[Normalize]=None
    ):
        if by_order:
            self._assign_group_colors_by_order(key, colormap)
        else:
            self._assign_group_colors_by_value(key, colormap, norm)

    def assign_group_labels(
            self,
            template_string: str,
            template_latex_label: Optional[str] = None
    ):
        for g in self.groups:
            g.label = template_string.format(**g.values)
            if template_latex_label:
                g.latex_label = template_latex_label.format(**g.values)
            else:
                g.latex_label = g.label

    def _assign_group_colors_by_order(self, key, colormap):
        sorted_values = sorted([g.values[key] for g in self.groups])
        for g in self.groups:
            g.color = colormap(sorted_values.index(g.values[key]))

    def _assign_group_colors_by_value(self, key, colormap, norm: Optional[Normalize]):
        if norm is None:
            raise RuntimeError("Norm has to be defined when using assign_group_colors by_value.")
        for g in self.groups:
            g.color = colormap(norm(g.values[key]))


class GroupIterator:

    def __init__(self,
                 data_reader_class: Type[DataReader],
                 index: Union[pd.Index, pd.MultiIndex],
                 values: Dict[str, Any],
                 simulation_folder_template: str,
                 data_file_suffix: str='data.h5',):
        self.data_reader_class = data_reader_class
        self.index = index
        self.values = values
        self._simulation_folder_template = simulation_folder_template
        self._data_file_suffix = data_file_suffix
        self.color = None
        self.label = None
        self.latex_label = None

    def __iter__(self):
        return self._generator()

    def __len__(self):
        return len(self.index)

    def _generator(self) -> Generator['Simulation', None, None]:
        for idx in self.index:
            path = self._simulation_folder_template.format(idx)
            dr = self.data_reader_class(os.path.join(path, self._data_file_suffix))
            yield Simulation(
                index=idx,
                path=path,
                data_reader=dr,
                analyser=Analyser(dr, os.path.join(path, 'analysis.h5'))
            )

    def has_matching_values(self, **kwargs):
        for key in kwargs:
            if key not in self.values:
                raise KeyError(f'Groups are not created with key {key}')
            if self.values[key] != kwargs[key]:
                return False
        return True

    def __repr__(self):
        return (f'GroupIterator: Index = {self.index}; '
                f'selected by values {self.values}')

    def __str__(self):
        return self.__repr__()

    def get_label_from_values(self) -> str:
        if len(self.values) is 0:
            return 'group_iterator_has_no_values'
        label = ''
        for key in self.values:
            v = self.values[key]
            label = label + key + '_' + str(v) + '_'
        return '_'.join(label[:-1].split())

    def __hash__(self):
        return hash(self.get_label_from_values())


@dataclass
class Simulation:
    index: int
    path: str
    data_reader: DataReader
    analyser: Analyser



from dataclasses import dataclass
from typing import Optional

import numba
import numpy as np

from actomyosin_analyser.analysis import geometry


def compute_msd(
        trajectory: np.ndarray,
        simulation_box: Optional[np.ndarray]=None
) -> 'MeanSquaredDisplacement':
    msd_array = _numba_compute_msd(trajectory, simulation_box)
    return MeanSquaredDisplacement.from_array(msd_array)


@numba.njit
def _numba_compute_msd(trajectory: np.ndarray, simulation_box: Optional[np.ndarray]) -> np.ndarray:
    lags = np.arange(1, len(trajectory))
    msd = np.empty((len(lags), 3))
    for lag in lags:
        if simulation_box is None:
            displacements = trajectory[lag:] - trajectory[:-lag]
        else:
            displacements = geometry.get_minimum_image_vector(trajectory[lag:], trajectory[:-lag], simulation_box)
        squared_summed = np.sum(displacements**2, axis=-1)
        msd[lag-1, 0] = lag
        msd[lag-1, 1] = np.nanmean(squared_summed)
        msd[lag-1, 2] = np.nanvar(squared_summed)
    return msd


@dataclass
class MeanSquaredDisplacement:
    lag: np.ndarray
    msd: np.ndarray
    variance_of_msd: np.ndarray
    lag_time: Optional[np.ndarray] = None

    def as_array(self) -> np.ndarray:
        a = np.empty((len(self.lag), 3))
        a[:, 0] = self.lag
        a[:, 1] = self.msd
        a[:, 2] = self.variance_of_msd
        return a

    @staticmethod
    def from_array(msd_array: np.ndarray) -> 'MeanSquaredDisplacement':
        return MeanSquaredDisplacement(
            msd_array[:, 0],
            msd_array[:, 1],
            msd_array[:, 2]
        )



from typing import Tuple, Union
import logging
import numpy as np
from scipy.optimize import leastsq
from scipy.special import gamma


def compute_G_advanced(lag_time: np.ndarray, msd: np.ndarray,
                       radius: float,
                       n_points_per_decade: int,
                       kT: float,
                       dimensionality: int, width: float) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute G for MSD with a linear-spaced log_lag_time.
    Based on the code of
    T. Maier, T. Haraszti (https://doi.org/10.1186/1752-153X-6-144).
    """

    log_lag_time, log_msd = get_logspaced_axes(lag_time, msd, n_points_per_decade)
    J = _compute_J(log_msd, radius, kT, dimensionality)

    J, log_lag_time = _get_valid_values(J, log_lag_time)

    N = len(log_lag_time)
    omega = np.full(N, np.nan)
    alpha = np.full(N, np.nan)
    g = np.full(N, np.nan)

    for i in range(N):
        alpha_i, g_i, om_i = _compute_g_i(J, i, log_lag_time, width)
        if g_i < 0 or not np.isfinite(g_i):
            print('invalid g_i for index i = {}'.format(i))
            continue
        g[i] = g_i
        omega[i] = om_i
        alpha[i] = alpha_i

    g2 = []
    omega2 = []

    for i in range(N):
        alphaI, alphaR, pre = _compute_alphaR_and_alphaI_and_pre(
            g, i, omega, width)

        g_i = g[i] * pre * complex(alphaR, alphaI)

        g2.append(g_i)
        omega2.append(omega[i])

    return np.array(omega2), np.array(g2)


def compute_G_danny_seara(
        tau: np.ndarray,
        msd: np.ndarray,
        bead_radius: float,
        dimensionality: int,
        kT: float,
        width: float,
        n_points_per_decade: int,
        clip: float=0.03
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Re-implementation of the matlab code on Danny Seara's github page
    https://github.com/dsseara/microrheology/blob/master/src/Moduli/calc_G.m .
    Originally code there was written by John C. Crocker and U. Penn.
    """

    tau, msd = get_logspaced_axes(tau, msd, n_points_per_decade)

    omega = 1/tau
    C = dimensionality*kT/(3*np.pi*bead_radius)
    pi_2_1 = np.pi/2 - 1

    m, d, dd = _logderive(tau, msd, width)

    G_laplace = C/((m * gamma(1+d)) * (1+(dd/2)))
    g, da, dda = _logderive(omega, G_laplace, width)

    alpha_real = np.cos((np.pi/2) * da) - pi_2_1 * da * dda
    alpha_imag = np.sin((np.pi/2) * da) - pi_2_1 * (1-da) * dda
    G_real = g * (1 / (1 + dda)) * alpha_real
    G_imag = g * (1 / (1 + dda)) * alpha_imag

    G_real[G_real < G_laplace * clip] = 0
    G_imag[G_imag < G_laplace * clip] = 0

    if np.abs(dd).max() > 0.15 or np.abs(dda).max() > 0.15:
        logging.warning("High curvature in data, moduli may be unreliable!")
    return omega, G_real + 1j * G_imag, dd, dda


def _logderive(
        x: np.ndarray,
        y: np.ndarray,
        width: float
):
    n_points = len(x)
    df = np.zeros(n_points)
    ddf = np.zeros(n_points)
    f2 = np.zeros(n_points)
    lx = np.log(x)
    ly = np.log(y)

    for i in range(n_points):
        w = np.exp(-(lx - lx[i])**2 / (2 * width**2))
        ww = w > 0.03
        p = np.polyfit(lx[ww], ly[ww], 2, w=w[ww])
        f2 = np.exp(p[2] + p[1] * lx[i] + p[0] * lx[i]**2)
        df[i] = p[1] + (2 * p[0] * lx[i])
        ddf[i] = 2 * p[0]

    return f2, df, ddf


def _compute_alphaR_and_alphaI_and_pre(g, i, omega, width):
    _pi_2 = np.pi / 2
    _pi_2_1 = _pi_2 - 1

    lo = np.log(omega)
    lg = np.log(g)
    w = np.exp(-(lo - lo[i])**2 / (2*width**2))
    ww = np.where(w > 0.03)
    fit = np.polyfit(lo[ww], lg[ww], 2, w=w[ww])
    c, b, a = fit
    beta = c
    alpha = b + 2 * c * np.log(omega[i])
    pre = 1.0 / (1 + beta)
    alphapi = _pi_2 * alpha
    alphaR = np.cos(alphapi - _pi_2_1 * alpha * beta)
    alphaI = np.sin(alphapi - _pi_2_1 * beta * (1 - alpha))
    return alphaI, alphaR, pre


def _compute_g_i(J, i, log_lag_time, width):

    lx = np.log(log_lag_time)
    w = np.exp(-(lx - lx[i])**2 / (2*width**2))
    ww = np.where(w > 0.03)
    coeffs = np.polyfit(lx[ww], np.log(J[ww]), 2, w=w[ww])
    a, b, c = coeffs
    beta = 2 * a
    alpha_i = b + 2 * a * np.log(log_lag_time[i])
    om_i = 1 / log_lag_time[i]
    yi = np.exp(c + b * lx[i] + a * lx[i]**2)
    g_i = 1 / (yi * gamma(1 + alpha_i))
    g_i = g_i / (1 + 0.5 * beta)
    return alpha_i, g_i, om_i


def _get_valid_values(J, log_lag_time):
    valid = (log_lag_time > 0) & np.isfinite(log_lag_time) & (J > 0) & np.isfinite(J)
    log_lag_time = log_lag_time[valid]
    J = J[valid]
    return J, log_lag_time


def get_logspaced_axes(
        t: np.ndarray,
        x: np.ndarray,
        n_points_per_decade: Union[int, float]
) -> Tuple[np.ndarray, np.ndarray]:
    if n_points_per_decade < 8:
        raise ValueError("number of points per decade has to be 8 or higher "
                         "for this method to work more reliably")
    ratio = t[-1]/t[0]
    powers = np.linspace(0, np.log10(ratio), int(n_points_per_decade * np.log10(ratio)) + 1)
    logspaced_t = 10**powers * t[0]
    logspaced_selection = np.searchsorted(t, logspaced_t)
    logspaced_selection = np.unique(logspaced_selection)
    logspaced_selection = logspaced_selection[logspaced_selection < len(t)]
    t = t[logspaced_selection]
    x = x[logspaced_selection]
    _validate_log_spacing_has_at_least_8_points_per_decade(t)
    return t, x


def _validate_log_spacing_has_at_least_8_points_per_decade(t: np.ndarray):
    """
    According to the code by Crocker and Penn that is on Danny Seara's
    github repository https://github.com/dsseara/microrheology/blob/master/src/Moduli/calc_G.m,
    we need at least 7-8 points per decade for this method to work.
    """
    ratio = t[-1]/t[0]
    powers = np.arange(0, int(np.log10(ratio)) + 1)
    decades = 10**powers
    edges = decades * t[0]
    for i in range(1, len(edges)):
        if ((t >= edges[i-1]) & (t < edges[i])).sum() < 8:
            raise RuntimeError("Need at least 8 data points per decade. The number of points"
                               " for your log-spacing is set too small.")


def _compute_J(msd, radius, kT, dim) -> np.ndarray:
    return 3/dim * np.pi * radius / kT * msd
